import os
import argparse
from pathlib import Path
from dotenv import load_dotenv
load_dotenv()

from .utils import *

def setup_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('-p', '--pipeline', type=str, help='specify the pipeline to run')
    parser.add_argument('-d', '--pipeline_dir', type=str, help='specify the pipeline dir to run')
    parser.add_argument('-i', '--input', type=str, help='specify input data')
    parser.add_argument('-o', '--output', type=str, help='specify output data')
    parser.add_argument('-m', '--mode', type=str, default='async', choices=['async', 'mp', 'seq'], help='specify the run mode')
    parser.add_argument('--llm', type=str, choices=['lmdeploy', 'vllm', 'mlx', 'ollama', 'openai', 'torch'], help='specify the llm backend')
    parser.add_argument('--model', type=str, help='specify the model path')
    parser.add_argument('--rag', type=str, choices=['json', 'http'], help='specify the rag backend')
    parser.add_argument('--split', type=int, help='split the data into parts to run')
    parser.add_argument('--png', action='store_true', help='export graph as png')
    parser.add_argument('--log', action='store_true', help='save logs')
    parser.add_argument('--test', action='store_true', help='run test')
    parser.add_argument('--serve', action='store_true', default=False, help='serve the pipeline as website & API')
    
    parser.add_argument('--env', action='store_true', default=False, help='run in environment mode, ignoring other required options')

    args, _ = parser.parse_known_args()

    if args.env:
        return args

    if not args.serve and not args.pipeline and not args.pipeline_dir:
        parser.error("the following arguments are required: -p/--pipeline or -d/--pipeline_dir")

    if args.log: os.environ['SAVE_LOG'] = '1'
    return args

def main():
    args = setup_args()

    if args.env:
        test_env()
        return

    if args.model:
        os.environ['MODEL_PATH'] = args.model

    from .manager import PipelineManager
    pm = PipelineManager(run_mode=args.mode, llm_type=args.llm, rag_type=args.rag, pipes_dir=args.pipeline_dir)

    if args.pipeline:
        pipefile = Path(args.pipeline)
        pipe = pm.add_pipe(pipefile.stem, pipefile=pipefile)

    if args.input:
        data = jload(args.input)
        r = pipe.run(data, split=args.split, save_perf=args.png)

        if args.output:
            if type(r) is tuple: jdump(r[0], args.output)
            elif type(r) is list: jdump([i for i,_ in r], args.output)
    elif args.png:
        pipe.to_png(f'{pipefile.stem}.png')
    elif args.serve:
        import uvicorn
        from .server import PipelineServer

        port = int(os.getenv('PORT', 8000))
        server = PipelineServer(pipeline_manager=pm)
        uvicorn.run(server.app, host="0.0.0.0", port=port, log_config=None, log_level=os.getenv('LOGGING_LEVEL', 'INFO').lower())



import json
import uuid
import time
import asyncio
import traceback
import threading
from pathlib import Path
from fastapi import APIRouter
from pydantic import BaseModel
from fastapi import HTTPException, WebSocket, WebSocketDisconnect
from fastapi.responses import JSONResponse, FileResponse
from .constant import *
from ..log import log
from .task import WSConnectionManager, TaskQueue, TaskWorker

class WorkspaceTaskQueue(TaskQueue):
    def queue_updated_broadcast(self):
        d = {"status": {"exec_info": self.get_queue_info()}}
        m = Message(Types.STATUS, d)
        self.loop.call_soon_threadsafe(self.ws_msges.put_nowait, m)

class WorkspaceTaskWorker(TaskWorker):
    def run(self):
        name = threading.current_thread().name

        while True:
            queue_item = self.queue.get(timeout=1000)
            if queue_item is not None:
                queue_id, (task_id, task_data, sid) = queue_item
                log.debug(f'{name}:\nqueue_id: {queue_id}\nsid: {sid}\ntask_id: {task_id}')

                try:
                    out = self.run_task(task_id, task_data, sid)
                except Exception as e:
                    err = traceback.format_exc()
                    log.error(err)
                    self.send_msg(Events.ERROR, {"error": err}, sid)
                    out = {"error": err}

                self.send_msg(Types.EXEC_SUCCESS, {"prompt_id": task_id}, sid)
                self.send_msg(Types.EXECUTING, {'prompt_id': task_id, 'node': None}, sid) # remove progress in the browser tab bar

                self.queue.task_done(
                    queue_id,
                    {
                        "outputs": out
                    },
                    status={
                        "status_str": 'success',
                        "completed": True,
                        "messages": None,
                    }
                )

    def run_task(self, task_id, task_data, sid):
        self.send_msg(Types.EXEC_START, {"prompt_id": task_id}, sid)

        d = {
            "prompt_id": task_id,
            "nodes": {
                1: {
                    'display_node_id': "1",
                    'state': "running",
                }
            }
        }
        self.send_msg(Types.PROG_STATE, d, sid)

        time.sleep(1)

        # node 2
        d = {
            "prompt_id": task_id,
            "nodes": {
                2: {
                    'display_node_id': "2",
                    'max': 20,
                    'state': "running",
                    'value': 0,
                },
            }
        }
        self.send_msg(Types.PROG_STATE, d, sid)

        for i in range(20):
            d = {
                "prompt_id": task_id,
                "nodes": {
                    2: {
                        'display_node_id': "2",
                        'max': 20,
                        'state': "running",
                        'value': i+1,
                    }
                }
            }
            self.send_msg(Types.PROG_STATE, d, sid)
            time.sleep(0.2)
        

        # node 3
        d = {
            "prompt_id": task_id,
            "nodes": {
                3: {
                    'display_node_id': "3",
                    'state': "running",
                },
            }
        }
        self.send_msg(Types.PROG_STATE, d, sid)

        d = {
            "prompt_id": task_id,
            "display_node": '3',
            'node': '3',
            "output": {
                'images': [
                    {
                        "filename": "ComfyUI_00005_.png",
                        "subfolder": "",
                        "type": "output"
                    }
                ]
            }
        }
        self.send_msg(Types.EXECUTED, d, sid)
        
        d = {
            "prompt_id": task_id,
            "nodes": {
                3: {
                    'display_node_id': "3",
                    'state': "finished",
                },
            }
        }
        self.send_msg(Types.PROG_STATE, d, sid)



        out = None
        # if self.pipeline_manager:
        #     msg_func = lambda out: self.send_msg(Events.TASK_ITEM_PROCESS, out, sid)
        #     def cancel_func(out):
        #         if self.queue.get_flags(reset=False).get(task_id, {}).get("cancel"):
        #             raise Exception(f"Task {task_id} cancelled during execution.")

        #     for i, task in enumerate(tqdm(task_data)):
        #         self.send_msg(Events.TASK_ITEM_START, {"task_index": i, "total": len(task_data)}, sid)
        #         pipe = self.pipeline_manager.pipes[task['pipe']]
        #         pipe.add_node_finish_callback(callbacks=[msg_func, cancel_func])
        #         out, info = pipe.run(task['data'])
        #         self.send_msg(Events.TASK_ITEM_DONE, out, sid)
        #         # self.send_msg(Events.TASK_ITEM_DONE, info, sid)

        return out

class WorkspaceAPI:
    def __init__(self, pipeline_manager):
        self.router = router = APIRouter(prefix='/workspace')
        ws_msges = asyncio.Queue()
        ws_manager = WSConnectionManager()
        task_queue = WorkspaceTaskQueue(ws_msges)

        async def ws_loop():
            while True:
                msg = await ws_msges.get()
                log.info(f'WS send: {msg}')
                await ws_manager.send(msg)

        @router.on_event("startup")
        async def startup_event():
            if task_queue.loop is None:
                log.debug("Setup Sigmaflow Workspace API")
                loop = asyncio.get_running_loop()
                task_queue.loop = loop
                WorkspaceTaskWorker(queue=task_queue, loop=loop, ws_msges=ws_msges, pipeline_manager=pipeline_manager).start()
                asyncio.create_task(ws_loop())

        @router.get("/api/users")
        async def users():
            try:
                ret = {
                    'storage': "server",
                    'migrated': True
                }
                return ret
            except Exception as e:
                raise HTTPException(status_code=500, detail=traceback.format_exc())
        
        @router.get("/api/i18n")
        async def i18n():
            try:
                return {}
            except Exception as e:
                raise HTTPException(status_code=500, detail=traceback.format_exc())
        
        @router.get("/api/system_stats")
        async def system_stats():
            try:
                return {
                    "system": {
                        "os": "posix",
                        "ram_total": 274877906944,
                        "ram_free": 270875971584,
                        "comfyui_version": "0.3.49",
                        "required_frontend_version": "1.24.4",
                        "python_version": "3.13.5 | packaged by Anaconda, Inc. | (main, Jun 12 2025, 16:09:02) [GCC 11.2.0]",
                        "pytorch_version": "2.7.1+cu128",
                        "embedded_python": False,
                        "argv": [
                            "main.py"
                        ]
                    },
                    "devices": [
                        {
                            "name": "cuda:0 NVIDIA A800-SXM4-80GB : cudaMallocAsync",
                            "type": "cuda",
                            "index": 0,
                            "vram_total": 17029083955,
                            "vram_free": 16594973491,
                            "torch_vram_total": 0,
                            "torch_vram_free": 0
                        }
                    ]
                }
            except Exception as e:
                raise HTTPException(status_code=500, detail=traceback.format_exc())

        @router.get("/api/settings")
        async def settings():
            try:
                return {
                    "Comfy.TutorialCompleted": True,
                    "Comfy.Release.Version": "0.3.44",
                    "Comfy.Release.Status": "what's new seen",
                    "Comfy.Release.Timestamp": 1752042448014,
                    "Comfy.ColorPalette": "dark",
                    "Comfy.Locale": "en"
                }
            except Exception as e:
                raise HTTPException(status_code=500, detail=traceback.format_exc())

        @router.get("/api/userdata")
        async def userdata():
            try:
                return []
            except Exception as e:
                raise HTTPException(status_code=500, detail=traceback.format_exc())
        
        @router.get("/api/extensions")
        async def extensions():
            try:
                return []
            except Exception as e:
                raise HTTPException(status_code=500, detail=traceback.format_exc())
        
        @router.get("/api/object_info")
        async def object_info():
            try:
                cur_folder = Path(__file__).parent
                with open(cur_folder / 'object_info.json', 'r') as f:
                    return json.load(f)
            except Exception as e:
                raise HTTPException(status_code=500, detail=traceback.format_exc())
        
        @router.get("/api/experiment/models")
        async def models():
            try:
                cur_folder = Path(__file__).parent
                with open(cur_folder / 'models.json', 'r') as f:
                    return json.load(f)
            except Exception as e:
                raise HTTPException(status_code=500, detail=traceback.format_exc())
        
        @router.get("/api/queue")
        async def queue():
            try:
                return {"queue_running": [], "queue_pending": []}
            except Exception as e:
                raise HTTPException(status_code=500, detail=traceback.format_exc())

        @router.get("/api/history")
        async def history(max_items: int):
            try:
                return {}
            except Exception as e:
                raise HTTPException(status_code=500, detail=traceback.format_exc())

        @router.get("/api/view")
        async def view(filename: str, subfolder: str):
            try:
                image_path = '/mnt/workspace/code/github/ComfyUI/output/ComfyUI_00001_.png'
                return FileResponse(image_path)
            except Exception as e:
                raise HTTPException(status_code=500, detail=traceback.format_exc())

        @router.post("/api/prompt")
        async def prompt(data: WorkspacePromptData):
            try:
                log.debug(f'prompt: {data.prompt}')
                prompt_id = str(data.prompt_id or uuid.uuid4())
                task_queue.put((prompt_id, data.prompt, data.client_id))
                response = {"prompt_id": prompt_id, "number": 1, "node_errors": {}}
                return response

                # node_errors = {
                #     4: {
                #         "errors": [{
                #             "type": "exception_during_validation",
                #             "message": "Exception when validating node",
                #             "details": 'test error',
                #             "extra_info": {
                #                 "exception_type": '',
                #                 "traceback": ''
                #             }
                #         }],
                #         "dependent_outputs": [],
                #         "class_type": "CheckpointLoaderSimple"
                #     }
                # }
                # return JSONResponse(status_code=400, content={"error": "test", "node_errors": node_errors})
            except Exception as e:
                raise HTTPException(status_code=500, detail=traceback.format_exc())

        @router.websocket("/ws")
        async def websocket_endpoint(ws: WebSocket, clientId: Optional[str] = None):
            sid = clientId or uuid.uuid4().hex
            await ws_manager.connect(ws, sid)

            try:
                data = {
                    'status': {
                        'exec_info':{
                            'queue_remaining': task_queue.get_tasks_remaining(),
                        }
                    },
                    'sid': sid,
                }
                m = Message(Types.STATUS, data, sid)
                await ws_manager.send(m)

                first_message = True
                while True:
                    data = await ws.receive()
                    match data["type"]:
                        case "websocket.receive":
                            if data["text"]:
                                ret = None
                                try:
                                    json_data = json.loads(data["text"])
                                except:
                                    log.error("Error parsing JSON")
                                    ret = {"error": "Invalid JSON format"}
                                    e = Events.ERROR
                                if ret is None and first_message and json_data.get("type") == "feature_flags":
                                    print(123)
                                    ret = {
                                        "max_upload_size": 104857600,
                                        "supports_preview_metadata": True
                                    }
                                    e = Types.FEATURE_FLAG
                                    first_message = False
                                if ret:
                                    m = Message(e, ret, sid)
                                    await ws_manager.send(m)
                        case "websocket.close":
                            break
                        case "websocket.disconnect":
                            break
            except WebSocketDisconnect:
                pass
            finally:
                ws_manager.disconnect(sid)

        @router.get("/internal/logs")
        async def logs():
            try:
                return {}
            except Exception as e:
                raise HTTPException(status_code=500, detail=traceback.format_exc())




import traceback
from fastapi import APIRouter, HTTPException
from .constant import *

class PipelineAPI:
    def __init__(self, pipeline_manager):
        self.router = router = APIRouter(prefix='/api')
        prompt_manager = pipeline_manager.prompt_manager

        @router.get("/list/{item}")
        async def list_item(item: str):
            try:
                match item:
                    case 'prompt':
                        ret = {k: {'text': v.text, 'keys': v.keys} for k,v in prompt_manager.prompts.items()}
                    case 'pipeline':
                        ret = pipeline_manager.export_pipe_conf()
                    case _:
                        raise HTTPException(status_code=400, detail="Invalid item type")
                return ret
            except Exception as e:
                raise HTTPException(status_code=500, detail=traceback.format_exc())

        @router.post("/update/{item}")
        async def update_item(item: str, p_data: PromptData | PipeData):
            try:
                match item:
                    case 'prompt':
                        if p_data.text:
                            prompt_manager.prompts[p_data.name].text = p_data.text
                        if p_data.keys:
                            prompt_manager.prompts[p_data.name].keys = p_data.keys

                        ret = {
                            p_data.name: {
                                'text': prompt_manager.prompts[p_data.name].text,
                                'keys': prompt_manager.prompts[p_data.name].keys
                            }
                        }
                    case 'pipeline':
                        ret = {
                            'result': pipeline_manager.update_pipe(p_data.name, p_data.data)
                        }
                    case _:
                        raise HTTPException(status_code=400, detail="Invalid item type")

                return ret
            except Exception as e:
                raise HTTPException(status_code=500, detail=traceback.format_exc())

        @router.get("/mermaid/{pipe_name}")
        async def mermaid(pipe_name: str):
            try:
                pipe = pipeline_manager.pipes[pipe_name]
                ret = {
                    'mermaid': pipe.pipetree.tree2mermaid(),
                }
                return ret
            except Exception as e:
                raise HTTPException(status_code=500, detail=traceback.format_exc())

        @router.post("/run/{pipe_name}")
        async def run_pipe(pipe_name: str, data: dict | list[dict]):
            try:
                pipe = pipeline_manager.pipes[pipe_name]
                ret = {
                    'result': await pipe.async_run(data)
                }
                return ret
            except Exception as e:
                raise HTTPException(status_code=500, detail=traceback.format_exc())




import os
import sys
import json
import shutil
import hashlib
import requests
import platform
import collections
from pathlib import Path, PosixPath

def check_cmd_exist(command):
    return shutil.which(command) is not None

def importpath(path):
    if type(path) is not PosixPath:
        strpath = str(path)
        if not strpath.startswith("/"):
            parent_path = Path(sys._getframe().f_globals.get("__file__", ".")).parent
            path = parent_path / path
        else:
            path = Path(path)

    try:
        sys.path.insert(0, str(path.parent))
        module = __import__(path.stem)
    finally:
        sys.path.pop(0)
    return module

def get_ordered_task(tasks):
    def get_dependencies(task_id, task_info):
        dependencies = []
        for key, value in task_info["inputs"].items():
            if isinstance(value, list):
                for item in value:
                    dep_task, _ = item.split("-")
                    dependencies.append(dep_task)
        return dependencies

    # Step 1: Build dependency graph and indegree count
    graph = collections.defaultdict(list)
    indegree = collections.defaultdict(int)

    for task_id, task_info in tasks.items():
        dependencies = get_dependencies(task_id, task_info)
        for dep in dependencies:
            graph[dep].append(task_id)
            indegree[task_id] += 1

    # Step 2: Topological sort using Kahn's algorithm
    execution_order = []
    queue = collections.deque([task_id for task_id in tasks if indegree[task_id] == 0])

    while queue:
        current = queue.popleft()
        execution_order.append(current)

        for neighbor in graph[current]:
            indegree[neighbor] -= 1
            if indegree[neighbor] == 0:
                queue.append(neighbor)

    return execution_order

def jload(inp):
    with open(inp, 'r', encoding='utf-8') as f:
        return json.load(f)

def jdump(obj, out):
    with open(out, 'w', encoding='utf-8') as f:
        if isinstance(obj, (dict, list)):
            json.dump(obj, f, indent=4, ensure_ascii=False)
        elif isinstance(obj, str):
            f.write(obj)
        else:
            raise ValueError(f"Unexpected type: {type(obj)}")

def calc_sha256(file_path):
    hash_sha256 = hashlib.sha256()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_sha256.update(chunk)
    return hash_sha256.hexdigest()

def get_latest_version(lib):
    from packaging import version
    package_name = lib[1]
    if package_name == "apex": return package_name, None
    if package_name == "swift": package_name = "ms-swift"
    url = f"https://pypi.org/pypi/{package_name}/json"
    response = requests.get(url)
    if response.status_code != 200: return package_name, None

    data = response.json()
    versions = data["releases"].keys()
    latest_version = max(versions, key=version.parse)
    return lib[1], latest_version

def test_env():
    from rich import print
    from rich.table import Table
    from rich.console import Console
    from rich.progress import Progress
    from concurrent.futures import ThreadPoolExecutor, as_completed

    console = Console()
    table = Table(title="LLM Environment Check Results")
    table.add_column("Library", style="cyan", justify="left")
    table.add_column("Status", style="green", justify="left")
    table.add_column("Details", style="white", justify="left")
    table.add_row('OS', "Installed", f"Platform: {platform.system()}, Release: {platform.release()}")
    table.add_row('Python', "Installed", "Version: " + sys.version.split()[0])

    if shutil.which("nvidia-smi"):
        driver_version = os.popen("nvidia-smi --version | grep DRIVER | awk '{print $4}'").read().strip()
        cuda_version = os.popen("nvidia-smi --version | grep CUDA | awk '{print $4}'").read().strip()
        table.add_row('Driver', "Installed", f"Version: {driver_version}")
        table.add_row('CUDA', "Installed", f"Version: {cuda_version}")

    try:
        import torch
        table.add_row('cuDNN', "Installed", f"Version: {torch.backends.cudnn.version()}")
        table.add_row('NCCL', "Installed", f"Version: {'.'.join(map(str,torch.cuda.nccl.version()))}")
    except ImportError:
        table.add_row('cuDNN', "[red]Not Installed[/red]")
        table.add_row('NCCL', "[red]Not Installed[/red]")

    if shutil.which("nvcc"):
        nvcc_version = os.popen("nvcc --version | grep release").read().split('release')[-1].strip()
        table.add_row('Toolkit', "Installed", f"Version: {nvcc_version}")
    table.add_section()

    libraries = [
        ("Apex", "apex"),
        ("PyTorch", "torch"),
        ("Transformer Engine", "transformer_engine"),
        ("FlashAttention", "flash_attn"),
        ("Transformers", "transformers"),
        ("DeepSpeed", "deepspeed"),
        ("Datasets", "datasets"),
        ("Tokenizers", "tokenizers"),
        ("vLLM", "vllm"),
        ("bitsandbytes", "bitsandbytes"),
        ("PEFT", "peft"),
        ("TRL", "trl"),
        ("wandb", "wandb"),
        ("lmdeploy", "lmdeploy"),
        ("Megatron-Core", "megatron.core"),
        ("ms-swift", "swift"),
        ("Triton", "triton"),
        ("ModelScope", "modelscope")
    ]

    versions_dict = {}
    with Progress() as progress:
        task = progress.add_task("[cyan]Checking latest versions...", total=len(libraries))

        with ThreadPoolExecutor(max_workers=8) as executor:
            future_to_lib = {executor.submit(get_latest_version, lib): lib for lib in libraries}
            for future in as_completed(future_to_lib):
                package_name, version = future.result()
                versions_dict[package_name] = version
                progress.update(task, advance=1)

    rows = []
    with Progress() as progress:
        task = progress.add_task("[cyan]Checking libraries...", total=len(libraries))
        for name, module in libraries:
            try:
                lib = __import__(module)
                status = "Installed"

                if module == "megatron.core":
                    ver = lib.core.__version__
                elif module == "apex":
                    ver = "unknown"
                else:
                    ver = lib.__version__

                if module == "apex":
                    details = "Apex is available"
                elif module == "torch":
                    details = f"Version: {ver}, CUDA: {lib.cuda.is_available()}"
                else:
                    details = f"Version: {ver}"
                
                v = versions_dict[module]
                if v and v != ver and not ver.startswith(v) and v > ver:
                    details += f"[red]  (Latest: {v})[/red]"
            except ImportError:
                status = "[red]Not Installed[/red]"
                details = "N/A"
            rows.append((name, status, details))
            progress.update(task, advance=1)

    rows.sort(key=lambda x: x[0])
    for row in rows: table.add_row(*row)
    table.add_section()

    try:
        import pynvml
        pynvml.nvmlInit()
        device_count = pynvml.nvmlDeviceGetCount()
        for i in range(device_count):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            name = pynvml.nvmlDeviceGetName(handle)
            meminfo = pynvml.nvmlDeviceGetMemoryInfo(handle)
            memstr = f'{int(meminfo.used/1024/1024/1024):d}G / {int(meminfo.total/1024/1024/1024):d}G, {meminfo.used/meminfo.total:3.0%}'
            temp = pynvml.nvmlDeviceGetTemperature(handle, 0)
            utlization = pynvml.nvmlDeviceGetUtilizationRates(handle)
            table.add_row('GPU' if not i else '', f'{i}', f"{name}, \\[mem] {memstr}, \\[utl] {utlization.gpu:3d}%, {temp:3d}°C")
        pynvml.nvmlShutdown()
    except ImportError:
        pass

    console.print(table)



from pathlib import Path
from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles

from .api import PipelineAPI
from .task import TaskAPI
from .workspace import WorkspaceAPI

class PipelineServer:
    def __init__(self, pipeline_manager=None):
        self.app = FastAPI(title='Sigmaflow Server')

        api = PipelineAPI(pipeline_manager)
        task = TaskAPI(pipeline_manager)
        workspace = WorkspaceAPI(pipeline_manager)

        self.app.include_router(api.router)
        self.app.include_router(task.router)
        self.app.include_router(workspace.router)

        web_root = Path(f'{__file__[:__file__.rindex("/")]}/website/dist/')
        work_root = Path(f'/mnt/workspace/code/github/ComfyUI_frontend/dist/')
        self.app.mount("/workspace/", StaticFiles(directory=work_root, html=True), name="SigmaFlow Workspace")
        self.app.mount("/", StaticFiles(directory=web_root, html=True), name="SigmaFlow Web")



from enum import Enum
from pydantic import BaseModel
from dataclasses import dataclass
from typing import Optional, Dict, Any, Callable

class Events(Enum):
    MSG = 'msg'
    PREVIEW_IMAGE = 'img'
    UNENCODED_PREVIEW_IMAGE = 'unencoded_img'
    WS_CONNECTED = 'ws_connected'
    TASK_START = 'task_start'
    TASK_ITEM_START = 'task_item_start'
    TASK_ITEM_PROCESS = 'task_item_process'
    TASK_ITEM_DONE = 'task_item_done'
    TASK_END = 'task_end'
    ERROR = 'error'

class Types(Enum):
    STATUS = 'status'
    FEATURE_FLAG = 'feature_flags'
    EXEC_START = 'execution_start'
    EXEC_CACHE = 'execution_cached'
    PROG_STATE = 'progress_state'
    EXECUTING = 'executing'
    PROGRESS = 'progress'
    EXECUTED = 'executed'
    EXEC_SUCCESS = 'execution_success'

@dataclass
class Message:
    header: Events | Types
    data: Any
    sid: Optional[str] = None

    def dict(self):
        m = {
            "event" if self.header is Events else "type": self.header.value,
            "data": self.data,
        }
        if self.sid: m['sid'] = self.sid
        return m

class BinaryEventTypes:
    PREVIEW_IMAGE = 1
    UNENCODED_PREVIEW_IMAGE = 2

class TaskData(BaseModel):
    sid: str | None = None
    task: list

class PromptData(BaseModel):
    name:   str | None = None
    text:   str | None = None
    keys:  list | None = None

class PipeData(BaseModel):
    name:   str | None = None
    data:  dict | None = None

class WorkspacePromptData(BaseModel):
    client_id: str | None = None
    extra_data: dict
    prompt: dict
    prompt_id: str | None = None



# from .prompt import Prompt
# from .pipeline import Pipeline
# from .manager import PromptManager, PipelineManager
# from .server import PipelineServer
# from .api import PipelineAPI
# from .log import log



import copy
import time
import uuid
import json
import heapq
import asyncio
import traceback
import threading
from tqdm.rich import tqdm
from starlette.websockets import WebSocketState
from fastapi import APIRouter, HTTPException, WebSocket, WebSocketDisconnect

from ..log import log
from .constant import *

class TaskQueue:
    def __init__(self, ws_msges, loop=None, max_history_size=10000):
        self.ws_msges = ws_msges
        self.loop = loop
        self.max_history_size = max_history_size
        self.mutex = threading.RLock()
        self.not_empty = threading.Condition(self.mutex)
        self.task_counter = 0
        self.queue = []
        self.currently_running = {}
        self.history = {}
        self.flags = {}

    def put(self, item):
        with self.mutex:
            heapq.heappush(self.queue, item)
            self.queue_updated_broadcast()
            self.not_empty.notify()

    def get(self, timeout=None):
        with self.not_empty:
            while len(self.queue) == 0:
                self.not_empty.wait(timeout=timeout)
                if timeout is not None and len(self.queue) == 0:
                    return None
            item = heapq.heappop(self.queue)
            task_id = self.task_counter
            self.currently_running[task_id] = copy.deepcopy(item)
            self.task_counter += 1
            self.queue_updated_broadcast()
            return task_id, item

    def task_done(self, task_id, history_result, status):
        with self.mutex:
            item = self.currently_running.pop(task_id)
            if len(self.history) > self.max_history_size:
                self.history.pop(next(iter(self.history)))

            self.history[item[0]] = {
                "task": item,
                'status': status,
            } | history_result
            self.queue_updated_broadcast()

    def queue_updated_broadcast(self):
        m = Message(Events.MSG, self.get_queue_info())
        self.loop.call_soon_threadsafe(self.ws_msges.put_nowait, m)

    def get_queue_info(self):
        return {'queue_remaining': self.get_tasks_remaining()}

    def get_current_queue(self):
        with self.mutex:
            out = []
            for x in self.currently_running.values():
                out += [x]
            return (out, copy.deepcopy(self.queue))

    def get_tasks_remaining(self):
        with self.mutex:
            return len(self.queue) + len(self.currently_running)

    def wipe_queue(self):
        with self.mutex:
            self.queue = []
            self.server.queue_updated()

    def delete_queue_item(self, function):
        with self.mutex:
            for x in range(len(self.queue)):
                if function(self.queue[x]):
                    if len(self.queue) == 1:
                        self.wipe_queue()
                    else:
                        self.queue.pop(x)
                        heapq.heapify(self.queue)
                    self.server.queue_updated()
                    return True
        return False

    def get_history(self, prompt_id=None, max_items=None, offset=-1):
        with self.mutex:
            if prompt_id is None:
                out = {}
                i = 0
                if offset < 0 and max_items is not None:
                    offset = len(self.history) - max_items
                for k in self.history:
                    if i >= offset:
                        out[k] = self.history[k]
                        if max_items is not None and len(out) >= max_items:
                            break
                    i += 1
                return out
            elif prompt_id in self.history:
                return {prompt_id: copy.deepcopy(self.history[prompt_id])}
            else:
                return {}

    def wipe_history(self):
        with self.mutex:
            self.history = {}

    def delete_history_item(self, id_to_delete):
        with self.mutex:
            self.history.pop(id_to_delete, None)

    def set_flag(self, name, data):
        with self.mutex:
            self.flags[name] = data
            self.not_empty.notify()

    def get_flags(self, reset=True):
        with self.mutex:
            if reset:
                ret = self.flags
                self.flags = {}
                return ret
            else:
                return self.flags.copy()

    def cancel_task(self, task_id):
        with self.mutex:
            if task_id in [v[0] for v in self.currently_running.values()]:
                self.flags[task_id] = {"cancel": True}
                return "running"
            for idx, item in enumerate(self.queue):
                if item[0] == task_id:
                    self.queue.pop(idx)
                    heapq.heapify(self.queue)
                    self.queue_updated_broadcast()
                    return "queued"
        return None

class TaskWorker(threading.Thread):
    def __init__(self, queue=None, loop=None, ws_msges=None, pipeline_manager=None):
        name = self.__class__.__name__
        threading.Thread.__init__(self, name=name, daemon=True)
        self.queue = queue
        self.loop = loop
        self.ws_msges = ws_msges
        self.pipeline_manager = pipeline_manager
        log.debug(f'{name} thread start')

    def run(self):
        name = threading.current_thread().name

        while True:
            queue_item = self.queue.get(timeout=1000)
            if queue_item is not None:
                queue_id, (task_id, task_data, sid) = queue_item
                log.debug(f'{name}:\nqueue_id: {queue_id}\nsid: {sid}\ntask_id: {task_id}')

                try:
                    out = self.run_task(task_id, task_data, sid)
                except Exception as e:
                    err = traceback.format_exc()
                    log.error(err)
                    self.send_msg(Events.ERROR, {"error": err}, sid)
                    out = {"error": err}

                self.queue.task_done(
                    queue_id,
                    {
                        "outputs": out
                    },
                    status={
                        "status_str": 'success',
                        "completed": True,
                        "messages": None,
                    }
                )
                self.send_msg(Events.TASK_END, {"task_id": task_id}, sid)

    def run_task(self, task_id, task_data, sid):
        self.send_msg(Events.TASK_START, {"task_id": task_id}, sid)

        out = None
        if self.pipeline_manager:
            msg_func = lambda out: self.send_msg(Events.TASK_ITEM_PROCESS, out, sid)
            def cancel_func(out):
                if self.queue.get_flags(reset=False).get(task_id, {}).get("cancel"):
                    raise Exception(f"Task {task_id} cancelled during execution.")

            for i, task in enumerate(tqdm(task_data)):
                self.send_msg(Events.TASK_ITEM_START, {"task_index": i, "total": len(task_data)}, sid)
                pipe = self.pipeline_manager.pipes[task['pipe']]
                pipe.add_node_finish_callback(callbacks=[msg_func, cancel_func])
                out, info = pipe.run(task['data'])
                self.send_msg(Events.TASK_ITEM_DONE, out, sid)
                # self.send_msg(Events.TASK_ITEM_DONE, info, sid)

        return out

    def send_msg(self, header, data, sid=None):
        data |= {"timestamp": int(time.time() * 1000)}
        m = Message(header, data, sid)
        self.loop.call_soon_threadsafe(self.ws_msges.put_nowait, m)

class WSConnectionManager:
    def __init__(self):
        self.active_connections: Dict[str, WebSocket] = {}

    async def connect(self, websocket: WebSocket, sid: str):
        await websocket.accept()
        self.active_connections[sid] = websocket

    def disconnect(self, sid: str):
        if sid in self.active_connections:
            del self.active_connections[sid]

    async def send(self, msg: Message):
        match msg.header:
            case Events.UNENCODED_PREVIEW_IMAGE:
                data = self.encode_image(msg.data)
                func = 'send_bytes'
            case Events.PREVIEW_IMAGE:
                data = msg.data
                func = 'send_bytes'
            case _:
                data = msg.dict()
                func = 'send_json'

        if msg.sid:
            websocket = self.active_connections.get(msg.sid)
            if websocket and websocket.application_state == WebSocketState.CONNECTED:
                await getattr(websocket, func)(data)
        else:
            for ws in self.active_connections.values():
                if ws.application_state == WebSocketState.CONNECTED:
                    await getattr(ws, func)(data)

    def encode_image(self, image_data):
        from PIL import Image, ImageOps
        image_type = image_data[0]
        image = image_data[1]
        max_size = image_data[2]
        if max_size is not None:
            resampling = Image.Resampling.BILINEAR if hasattr(Image, 'Resampling') else Image.ANTIALIAS
            image = ImageOps.contain(image, (max_size, max_size), resampling)
        type_num = 1 if image_type == "JPEG" else 2
        bytes_io = BytesIO()
        header = struct.pack(">I", type_num)
        bytes_io.write(header)
        image.save(bytes_io, format=image_type, quality=95, compress_level=1)
        preview_bytes = bytes_io.getvalue()
        return preview_bytes

class TaskAPI:
    def __init__(self, pipeline_manager):
        self.router = router = APIRouter()
        ws_msges = asyncio.Queue()
        ws_manager = WSConnectionManager()
        task_queue = TaskQueue(ws_msges)

        async def ws_loop():
            while True:
                msg = await ws_msges.get()
                log.info(f'WS send: {msg}')
                await ws_manager.send(msg)

        @router.on_event("startup")
        async def startup_event():
            if task_queue.loop is None:
                log.debug("Setup Sigmaflow Task API")
                loop = asyncio.get_running_loop()
                task_queue.loop = loop
                TaskWorker(queue=task_queue, loop=loop, ws_msges=ws_msges, pipeline_manager=pipeline_manager).start()
                asyncio.create_task(ws_loop())

        @router.get("/task")
        async def get_task():
            return task_queue.get_queue_info()

        @router.get("/cur_task")
        async def get_cur_task():
            running, pending = task_queue.get_current_queue()
            queue_info = {
                'queue_running': running,
                'queue_pending': pending,
            }
            return queue_info

        @router.get("/cancel_task/{task_id}")
        async def cancel_task(task_id: str):
            status = task_queue.cancel_task(task_id)
            if status == "queued":
                return {"status": "removed_from_queue", "task_id": task_id}
            elif status == "running":
                return {"status": "cancelling_running_task", "task_id": task_id}
            else:
                raise HTTPException(status_code=404, detail="Task not found")

        @router.post("/task")
        async def process_task(data: TaskData):
            try:
                task_id = str(uuid.uuid4())
                task_queue.put((task_id, data.task, data.sid))

                ret = {"task_id": task_id}
                return ret
            except Exception as e:
                print(traceback.format_exc())
                raise HTTPException(status_code=500, detail=traceback.format_exc())

        @router.websocket("/ws")
        async def websocket_endpoint(ws: WebSocket, sid: Optional[str] = None):
            sid = sid or uuid.uuid4().hex
            await ws_manager.connect(ws, sid)

            try:
                data = {
                    'queue_remaining': task_queue.get_tasks_remaining(),
                }
                m = Message(Events.WS_CONNECTED, data, sid)
                await ws_manager.send(m)

                while True:
                    data = await ws.receive()
                    match data["type"]:
                        case "websocket.receive":
                            if data["text"]:
                                ret = None
                                try:
                                    task_data = json.loads(data["text"])
                                except:
                                    log.error("Error parsing JSON")
                                    ret = {"error": "Invalid JSON format"}
                                    e = Events.ERROR
                                if ret is None:
                                    task_id = str(uuid.uuid4())
                                    task_queue.put((task_id, task_data["task"], sid))
                                    ret = {"task_id": task_id}
                                    e = Events.MSG
                                m = Message(e, ret, sid)
                                await ws_manager.send(m)
                        case "websocket.close":
                            break
                        case "websocket.disconnect":
                            break
            except WebSocketDisconnect:
                pass
            finally:
                ws_manager.disconnect(sid)
